/*
 *  linux/fs/namei.c
 *
 *  Copyright (C) 1991, 1992  Linus Torvalds
 */
/*
 *  $Id: namei.c,v 1.6 2005/01/09 21:43:25 ezk Exp $
 */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif /* HAVE_CONFIG_H */
#include "fist.h"
#include "unionfs.h"
#include <linux/quotaops.h>
#include <linux/dnotify.h>

/* This file contains functions lifted from Linux/fs/namei.c. The only reason
 * we have it is to support vfs_unlink_nozombie. */


/*
 * It's inline, so penalty for filesystems that don't use sticky bit is
 * minimal.
 */
static inline int check_sticky(struct inode *dir, struct inode *inode)
{
	if (!(dir->i_mode & S_ISVTX))
		return 0;
	if (inode->i_uid == current->fsuid)
		return 0;
	if (dir->i_uid == current->fsuid)
		return 0;
	return !capable(CAP_FOWNER);
}

/*
 *	Check whether we can remove a link victim from directory dir, check
 *  whether the type of victim is right.
 *  1. We can't do it if dir is read-only (done in permission())
 *  2. We should have write and exec permissions on dir
 *  3. We can't remove anything from append-only dir
 *  4. We can't do anything with immutable dir (done in permission())
 *  5. If the sticky bit on dir is set we should either
 *	a. be owner of dir, or
 *	b. be owner of victim, or
 *	c. have CAP_FOWNER capability
 *  6. If the victim is append-only or immutable we can't do antyhing with
 *     links pointing to it.
 *  7. If we were asked to remove a directory and victim isn't one - ENOTDIR.
 *  8. If we were asked to remove a non-directory and victim isn't one - EISDIR.
 *  9. We can't remove a root or mountpoint.
 */
static inline int may_delete(struct inode *dir,struct dentry *victim, int isdir)
{
	int error;

	if (!victim->d_inode || victim->d_parent->d_inode != dir)
		return -ENOENT;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
	error = permission(dir,MAY_WRITE | MAY_EXEC);
#else
	error = permission(dir, MAY_WRITE | MAY_EXEC, NULL);
#endif
	if (error)
		return error;
	if (IS_APPEND(dir))
		return -EPERM;
	if (check_sticky(dir, victim->d_inode)||IS_APPEND(victim->d_inode)||
	    IS_IMMUTABLE(victim->d_inode))
		return -EPERM;
	if (isdir) {
		if (!S_ISDIR(victim->d_inode->i_mode))
			return -ENOTDIR;
		if (IS_ROOT(victim))
			return -EBUSY;
	} else if (S_ISDIR(victim->d_inode->i_mode))
		return -EISDIR;
	if (IS_DEADDIR(dir))
		return -ENOENT;
	return 0;
}

/*	Check whether we can create an object with dentry child in directory
 *  dir.
 *  1. We can't do it if child already exists (open has special treatment for
 *     this case, but since we are inlined it's OK)
 *  2. We can't do it if dir is read-only (done in permission())
 *  3. We should have write and exec permissions on dir
 *  4. We can't do it if dir is immutable (done in permission())
 */
static inline int
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
may_create(struct inode *dir, struct dentry *child)
#else
may_create(struct inode *dir, struct dentry *child, struct nameidata *nd)
#endif
{
	if (child->d_inode)
		return -EEXIST;
	if (IS_DEADDIR(dir))
		return -ENOENT;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
	return permission(dir,MAY_WRITE | MAY_EXEC);
#else
	return permission(dir,MAY_WRITE | MAY_EXEC, nd);
#endif
}

/* Lifted from Linux/fs/namei.c:1469, but will not down dir->i_zombie. */
int vfs_unlink_nozombie(struct inode *dir, struct dentry *dentry)
{
	int error;

	print_entry_location();

	/* REMOVED: down(&dir->i_zombie); */
	error = may_delete(dir, dentry, 0);
	if (!error) {
		error = -EPERM;
		if (dir->i_op && dir->i_op->unlink) {
			DQUOT_INIT(dir);
			if (d_mountpoint(dentry))
				error = -EBUSY;
			else {
				lock_kernel();
				error = dir->i_op->unlink(dir, dentry);
				unlock_kernel();
				if (!error)
					d_delete(dentry);
			}
		}
	}
	/* REMOVED: up(&dir->i_zombie); */

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,20)
	if (!error)
		inode_dir_notify(dir, DN_DELETE);
#endif
	print_exit_status(error);
	return error;
}
//DQ: vfs_create now also takes a nameidata struct we might want to make this one take one also.
/* Lifted from Linux/fs/namei.c:960, but will not down dir->i_zombie. */
int vfs_create_nozombie(struct inode *dir, struct dentry *dentry, int mode)
{
	int error;

	mode &= S_IALLUGO;
	mode |= S_IFREG;

	/* REMOVED: down(&dir->i_zombie); */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
	error = may_create(dir, dentry);
#else
	//DQ: the null gets changed to a nameidata struct if we choose to add it here.
	error = may_create(dir, dentry,NULL);
#endif
	if (error)
		goto exit_lock;

	error = -EACCES;	/* shouldn't it be ENOSYS? */
	if (!dir->i_op || !dir->i_op->create)
		goto exit_lock;

	DQUOT_INIT(dir);
	lock_kernel();
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
	error = dir->i_op->create(dir, dentry, mode);
#else
	error = dir->i_op->create(dir, dentry, mode, NULL);
#endif
	unlock_kernel();
exit_lock:
	/* REMOVED: up(&dir->i_zombie); */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,20)
	if (!error)
		inode_dir_notify(dir, DN_CREATE);
#endif
	return error;
}
